import re
import json
import time
import logging
import subprocess
import ipaddress
import sys
import os
from typing import List, Optional, Tuple

import pytest

# Add the root directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
from src.device import HostDevice
from src.connection import TelnetConnection

logger = logging.getLogger(__name__)


class TestCollector:
    """ A pytest plugin to collect test items. """
    def pytest_collection_finish(self, session):
        """ Called after test collection has been completed and modified. """
        self.test_items = session.items


def get_tests() -> TestCollector:
    collector = TestCollector()
    pytest.main(["--no-header", "--no-summary", "-qq", "--collect-only"], plugins=[collector])
    test_items = collector.test_items
    return test_items


def load_json(file_path):
    # Assuming that the file is JSON
    with open(file_path, 'r') as json_file:
        data = json.load(json_file)
    return data


def print_banner(*messages: str, banner_legth = 80) -> None:
    """Prints a banner out of any number of messages."""
    border = "*" * banner_legth
    logger.info(border)
    for message in messages:
        logger.info(message.center(banner_legth))
    logger.info(border)


def execute_shell_commands_on_host(commands: List[str], print_response = False, quiet = False) -> Optional[str]:
    # Might require root privileges
    responses = []
    for command in commands:
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        if process.returncode != 0 and not quiet:
            logger.error(f"Error executing command: {command}")
            logger.error(f"Error message: {stderr.decode()}")
        else:
            response = stdout.decode()
            if print_response and not quiet:
                logger.debug(f"Command executed successfully: {command}")
                logger.debug(f"Output: {response}")
        if stdout:
            responses.append(stdout.decode())
        if stderr:
            responses.append(stderr.decode())
    if responses:
        return '\n'.join(responses)
    else:
        return None


def set_interface_ip(interface_name: str, ip: str, password: str, netmask: str = "24") -> None:
    if '/' not in ip:
        ip = f"{ip}/{netmask}"
    command = f"echo {password} | sudo -S ip addr add {ip} dev {interface_name}"
    execute_shell_commands_on_host([command], quiet=True)


def del_interface_ip(interface: str, ip: str, password: str, netmask: str = "24") -> None:
    if '/' not in ip:
        ip = f"{ip}/{netmask}"
    command = f"echo {password} | sudo -S ip addr del {ip} dev {interface}"

    # If the IP to be deleted is the only one on the interface, skip
    # Useful for: if developer uses their standard IP, it will not be deleted
    ipv4s, ipv6s = get_interface_ips(interface)
    if len(ipv4s) == 1 and ip.split('/')[0] in ipv4s:
        return
    if len(ipv6s) == 1 and ip.split('/')[0] in ipv6s:
        return
    execute_shell_commands_on_host([command])


def get_interface_ips(interface: str) -> Tuple[List[str], List[str]]:
    response = execute_shell_commands_on_host([f"ip addr show {interface}"])
    ipv4_pattern = r"\binet (\b\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\b)"
    ipv4_matches = re.findall(ipv4_pattern, response)
    ipv6_pattern = r"\binet6 ([a-f0-9:]+)"
    ipv6_matches = re.findall(ipv6_pattern, response)
    return ipv4_matches if ipv4_matches else [], ipv6_matches if ipv6_matches else []


def reboot_device(connection: "TelnetConnection", timeout: int = 60) -> "TelnetConnection":
    if connection is None or not connection.is_connected():
        raise ConnectionError("Connection is not established. Cannot reboot device.")

    vm_ip = connection.destination_ip
    vm = connection.destination_device
    connection.write_command("show expert system command bash")
    connection.write_command("/sbin/reboot")
    connection.disconnect()

    start_time = time.time()
    while True:
        packet_loss = get_packet_loss(vm_ip)
        if packet_loss == 0:
            break
        if timeout and time.time() - start_time > timeout:
            raise TimeoutError(f"Rebooting device {vm} took too long. Timeout reached.")

    connection.connect(vm, vm_ip)
    return connection


def ping(destination_ip: str, count: int = 1) -> str:
    return execute_shell_commands_on_host([f"ping -c {count} {destination_ip}"])


def get_packet_loss(response: str) -> str:
    """
    Example of result line in response:
        '2 packets transmitted, 2 received, 0% packet loss, time 11ms'
    Returned value:
        '0' (if pings have been successful)
    """
    match = re.search(r"\d+(?=%)", response)
    if match:
        return match.group()
    else:
        logger.critical(f"Ping: Could not find packet loss percentage in response: {response}")


def is_valid_ip(ip: str) -> bool:
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False


def scp_file_to_home_dir(local_file_path: str, user_at_ip: str, password: str) -> None:
    # Check if sshpass is installed on device
    host_vm = HostDevice()
    response = host_vm.write_command("sshpass")
    # response = execute_shell_commands_on_host(["sshpass"])
    if "Usage: sshpass" not in response:
        logger.critical('sshpass is not installed on the device. Please install it by "sudo apt install sshpass"')
    command = f"sshpass -p {password} scp {local_file_path} {user_at_ip}:~"
    response = host_vm.write_command(command)
    if response is None:
        return
    elif "No such file or directory" in response:
        logger.critical(f"Some file path is not valid. Local: {local_file_path}, @: {user_at_ip}")
        raise FileNotFoundError
    else:
        logger.critical(f"Unknown error occurred while copying file to the device. Got response: {response}")
        return ValueError



# connection.py


"""
This module contains the implementation for establishing and managing
    network connections between devices.
    Connection class contains the core actions of the connection,
    while TelnetConnection and TelnetCLIConnection have to do with how the connection is established.

TelnetConnection: Connection originating from the host device to another device.
                  Creates, initialises and uses a Telnet object from telnetlib library.

TelnetCLIConnection: Connection originating from an already connected device to another device.
                     Uses the Telnet object from the TelnetConnection to connect to the next device.
                     This class requires that already-connected Telnet object
                         in order to be instantiated and used.
                     If established, the base connection is not available for use by other connections,
                         until the TelnetCLIConnection is closed.
"""


import logging
import re
import socket
import time
import sys
import os
from abc import ABC, abstractmethod
from typing import Optional, List, Union

import telnetlib

# Add the root directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
from src.device import Device


logger = logging.getLogger(__name__)


class Connection(ABC):
    def __init__(self, timeout: int=10):
        self.destination_device = None
        self.destination_ip = None
        self.resulting_telnet_connection = None
        self.timeout = timeout
        self._is_occupied = False  # Signifies if the connection is in use by another connection
        self.prompt_symbol = None

    @abstractmethod
    def connect():
        pass

    @abstractmethod
    def disconnect():
        pass

    def check_occupied(func):
        """
        Decorator to check if the connection is already in use and hence not available.

        This decorator is used to wrap methods that should not be executed if the connection is already in use.

        Raises:
            ConnectionRefusedError: If the connection is already in use.
        """
        def wrapper(self, *args, **kwargs):
            if self._is_occupied:
                raise ConnectionRefusedError("This connection is already in use. Please close the connections that use it first.")
            return func(self, *args, **kwargs)
        return wrapper

    def check_device_type(required_type, is_root: bool=False):
        """
        Decorator to check the device type and connection privileges before executing a function.

        Args:
            required_type (str): The required device type for the function to be executed.
            is_root (bool, optional): If True, the function requires root privileges to be executed. Defaults to False.

        Raises:
            ValueError: If the device is not of the required type.
            ConnectionError: If the device is not connected.
            PermissionError: If root privileges are required but the user does not have them.
        """
        def decorator(func):
            def wrapper(self, *args, **kwargs):
                if self.destination_device.type != required_type:
                    raise ValueError(f'This method is available only for {required_type} devices, but the destination device is of type "{self.destination_device.type}".')
                # Perform the connection check too, since it's a common requirement
                if not self.is_connected:
                    raise ConnectionError("Device is not connected")
                if is_root and not self.is_root:
                    raise PermissionError("Root privileges required to perform this action")
                return func(self, *args, **kwargs)
            return wrapper
        return decorator

    def check_connection(func):
        """
        Decorator to check if the device is connected before executing a function.

        Raises:
            ConnectionError: If the device is not connected.
        """
        def wrapper(self, *args, **kwargs):
            if not self.is_connected:
                raise ConnectionError("Device is not connected")
            return func(self, *args, **kwargs)
        return wrapper

    @check_occupied
    def write_command(self, command: str, expected_prompt_pattern: Optional[List[str]]=None, timeout: Optional[int]=None) -> Optional[str]:
        """
        Writes a command to the telnet connection and returns the response.

        This method sends a command to the device via the telnet connection, waits for a response, and then returns that response.
        The response is expected to end with a prompt symbol or match an expected pattern, which is specified by the `expected_prompt_pattern` parameter.

        Args:
            command (str): The command to be sent to the device.
            expected_prompt_pattern (Optional[List[str]]): A list of regex patterns that the response is expected to match. If None, the method waits for the prompt symbol. Defaults to None.
            timeout (Optional[int]): The maximum time to wait for a response, in seconds. If None, the method uses the default timeout. Defaults to None.

        Returns:
            Optional[str]: The response from the device, or None if there was no response.

        Raises:
            ConnectionError: If the telnet connection is not established.
        """
        self.flush()  # Make sure nothing is in the buffer

        if self.resulting_telnet_connection is not None:
            # If the command is a string, encode it to bytes first
            command = command.encode("ascii") + b"\r" if hasattr(command, 'encode') else command
            self.resulting_telnet_connection.write(command)
            assert self.prompt_symbol is not None, "Prompt symbol is not defined."

            # "expect" can wait for multiple patterns
            if expected_prompt_pattern:
                response = self.resulting_telnet_connection.expect(
                    expected_prompt_pattern, timeout or self.timeout,
                )[2]  # The third element of the tuple is the response
            # but "read_until", while only for one pattern (prompt_symbol), is more reliable
            else:
                response = self.resulting_telnet_connection.read_until(
                    self.prompt_symbol.encode("ascii"), timeout or self.timeout
                )
            response = response.decode('ascii') if response else None
        else:
            raise ConnectionError("No connection object from Telnet found during write_command.")
        return response

    @check_occupied
    def flush(self, time_interval: int = 0.1) -> None:
        """
        This method waits for a short period of time to allow any remaining data to arrive,
        then reads and discards all data that has arrived at the telnet connection.
        """
        try:
            time.sleep(time_interval)
            if self.resulting_telnet_connection is not None:
                self.resulting_telnet_connection.read_very_eager()
        except EOFError as eof:
            logger.error(f"EOFError. Usually something is wrong while loading the connection. | {eof}")
            raise EOFError

    @check_occupied
    def flush_deep(self, time_interval: int = 0.1, retries_timeout: int = 60) -> None:
        logger.debug("Deep flushing ...")
        end_pattern = f"{self.prompt_symbol}"
        if retries_timeout > 0:
            start_time = time.time()
        while True:
            response = self.write_command('\n', timeout=time_interval)
            if response is not None and end_pattern in response.strip():
                break
            if retries_timeout > 0 and time.time() - start_time > retries_timeout:
                raise TimeoutError("Timeout while flushing deep")

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        pass

    @check_occupied
    def read_until(self, prompt: bytes, timeout: Optional[int] = None) -> Optional[str]:
        """
        Reads data from the telnet connection until a specified prompt is encountered or until timeout.

        Args:
            prompt (bytes): The prompt to read until.
            timeout (Optional[int]): The maximum time to wait for the prompt, in seconds. If None, the method uses the default timeout. Defaults to None.

        Returns:
            Optional[str]: The data read from the connection, or None if no data was read.
        """
        if timeout is not None:
            self.timeout = timeout
        if self.resulting_telnet_connection is not None:
            response = self.resulting_telnet_connection.read_until(
                prompt, self.timeout
            )
            response = response.decode('ascii') if response else None
        else:
            raise NotImplementedError("No connection object from Telnet found during read_until.")
        return response

    @check_device_type("oneos")
    def load_config(self, config_path: str) -> None:
        """
        Loads a configuration file to a OneOS device.

        Args:
            config_path (str): The path to the configuration file.

        Raises:
            ValueError: If the device is not a OneOS device.
            OSError: If the configuration file fails to open (might not exist).
            ConnectionError: If the device is not connected.
        """
        logger.debug(f"Loading config {config_path.split('/')[-1]} ...")
        self.write_command("term len 0")
        with open(config_path) as fp:
            for line in fp:
                if line.strip().startswith("!"):
                    continue  # Skip comment lines
                if "hostname" in line:
                    self.destination_device.hostname = line.split()[-1]
                response = self.write_command(line)

        # Check that prompt has exited config terminal fully. Search for "localhost#" (default) or "<configured_hostname>#"
        self.prompt_symbol = f"{self.destination_device.hostname}#"
        response = self.write_command('\n').strip()
        if response != self.prompt_symbol:
            logger.warning(f"Loading config might have failed, prompt is not as expected. Received {response} but expected {self.prompt_symbol} instead")
            logger.debug('Sometimes the developer has miscalculated the "exit" commands in the BSA')
            self.write_command("end")
        logger.info(f"Loaded configuration to device {self.destination_device.hostname}")

    @check_device_type("oneos")
    def patch_config(self, config_path: str) -> None:
        logger.debug(f"Patching config {config_path.split('/')[-1]} ...")
        # If it has beed set as <hostname><prompt_symbol>, just keep the <prompt_symbol>
        # That is to avoid looking for "localhost#" but getting "localhost(config)#" during reconfig
        if len(self.prompt_symbol) != 1:
            self.prompt_symbol = self.prompt_symbol[-1]
        self.load_config(config_path)


    @check_device_type("linux")
    def set_sudo(self, root_password: Optional[str] = None) -> None:
        """
        Sets sudo privileges for a Linux device.
        The prompt symbol is updated to '#' to reflect the change to the root user.

        Args:
            root_password (Optional[str]): The root password. If None, the method uses the password of the destination device. Defaults to None.

        Raises:
            ValueError: If the device is not a Linux device.
            ConnectionError: If the device is not connected.
            AssertionError: If the method fails to switch to the root user.
        """
        if root_password is None:
            root_password = self.destination_device.password
        self.write_command("sudo su", expected_prompt_pattern=[b'password for user:'])
        self.write_command(root_password, expected_prompt_pattern=[b'#'])
        self.prompt_symbol = '#'  # In Linux, changes from '$' to '#' if root
        assert self.is_root, "Failed to identify root user"
        logger.info(f"Sudo privileges set for linux device: {self.destination_device.hostname}")

    @property
    def is_root(self) -> bool:
        """
        Checks if the current user is root on a Linux device by writing the 'whoami' command and checking the response.

        Returns:
            bool: True if the current user is root, False otherwise.
        """
        user = self.write_command("whoami", [b'\$', b'#']).split()[1].strip()
        return user == "root"

    @check_device_type("linux", is_root=True)
    def set_interface_ip(self, interface_name: str, ip_addr: str, netmask: str = "24", interface_state: str = "up") -> None:
        """
        Sets the IP address, netmask, and state of a specified interface on a Linux device.

        Args:
            interface_name (str): The name of the interface.
            ip_addr (str): The IP address to set.
            netmask (str, optional): The netmask to set. Defaults to "24".
            interface_state (str, optional): The state of the interface. Must be 'up' or 'down'. Defaults to "up".

        Raises:
            ValueError: If the IP address is invalid, if the interface does not exist or if the device type is not Linux.
            ConnectionError: If the device is not connected.
            PermissionError: If the user does not have root privileges.
        """
        if not re.match(r"\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b", ip_addr):
            raise ValueError("Invalid IP address.")
        if interface_state not in ["up", "down"]:
            logger.error(f"Invalid state: {interface_state}. Must be 'up' or 'down'. Passing 'up' by default.")
            interface_state = "up"
        if self._get_interface(interface_name) is None:
            raise ValueError(f"Interface {interface_name} not found")
        self.write_command(f"ip addr add {ip_addr}/{netmask} dev {interface_name}")
        self.write_command(f"ip link set {interface_name} {interface_state}")
        logger.info(f"Interface {interface_name} set to IP {ip_addr} with netmask {netmask} and state {interface_state}")

    @check_device_type("linux", is_root=True)
    def delete_interface_ip(self, interface_name: str, ip_addr: str, netmask: str = "24") -> None:
        """
        Deletes the IP address from a specified interface on a Linux device.

        Args:
            interface_name (str): The name of the interface.
            ip_addr (str): The IP address to delete.
            netmask (str, optional): The netmask of the IP address. Defaults to "24".

        Raises:
            ValueError: If the IP address is invalid, if the interface does not exist or if the device type is not Linux.
            ConnectionError: If the device is not connected.
            PermissionError: If the user does not have root privileges.
        """
        if not re.match(r"\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b", ip_addr):
            raise ValueError("Invalid IP address.")
        if self._get_interface(interface_name) is None:
            raise ValueError(f"Interface {interface_name} not found")
        self.write_command(f"ip addr del {ip_addr}/{netmask} dev {interface_name}")
        logger.info(f"IP {ip_addr} with netmask {netmask} deleted from interface {interface_name}")

    def _get_interfaces(self) -> Optional[List[List[str]]]:
        """
        Gets a list of all interfaces on the device.
        """
        interfaces = re.split(r'\r\n(?=\d)', self.write_command("ip a"))[1:]  # Disregard the CLI command
        return interfaces

    def _get_interface(self, interface_name: str):
        """
        Gets information about a specific interface on the device.
        """
        interface_list = self._get_interfaces()
        for interface in interface_list:
            if interface_name in interface:
                return interface
        return None

    @check_device_type("oneos")
    def unload_interface(self, interface_line: str, wrap_command: bool = True) -> Optional[str]:
        """
        Resets the configuration of a specified interface to its default settings.
        OneOS6 WARNING: "By configuring the interface back to default, it is possible that some services will not work any more"

        Args:
            interface_line (str): The line of the full interface name to reset (i.e. interface gigabitethernet 0/0).
            wrap_command (bool, optional): If True, the method enters and exits the "configure terminal" command.

        Returns:
            Optional[str]: The response from the device after sending the 'default' command, or None if there was no response.
        """
        self.write_command("config terminal") if wrap_command else None
        response = self.write_command(f"default {interface_line}")
        self.write_command("end") if wrap_command else None
        return response

    @check_device_type("oneos")
    def unload_config(self, unload_specific_commands: Optional[List[str]] = None, check_is_empty: bool = False) -> None:
        """
        Unloads the configuration of the device using a bottom-up approach.
        The configurations on the bottom of the config inherit properties from the configurations above them.

        Sometimes, even by that approach, some commands cannot be unloaded. In that case, the user must manually unload them,
            by providing the no-commands in the unload_specific_commands parameter.

        The config is retrieved by the very slow "show running-config" command. If check_is_empty is True,
            "show running-config" is called again (another couple of seconds wait time), that's why default is to not check.

        Args:
            unload_specific_commands (Optional[List[str]]): A list of specific commands to unload. Defaults to None.
            check_is_empty (bool, optional): If True, the method checks if the configuration is empty after unloading. Defaults to False.

        Raises:
            ValueError: If the configuration is not fully unloaded and check_is_empty is True, or if device type is not oneos.
            ConnectionError: If the device is not connected.
        """
        logger.debug(f"Unloading config for device {self.destination_device.hostname} ...")
        self.write_command("term len 0")
        self.flush()

        config_lines = self.write_command("show running-config").split("\n")
        config_lines_reverse = config_lines[::-1]  # Traverse from bottom to top

        self.prompt_symbol = "#"
        self.write_command("config terminal")

        # Unload ip routes
        for line in config_lines_reverse:
            if re.search(r"^(ip(v6|) (route|host)|aaa authentication login)", line):
                self.write_command(f"no {line}")
            elif re.search(r"^radius-server", line):
                self.write_command(f"no radius-server {line.split(' ')[1]}")
            if "exit" in line:
                break


        # Unload interfaces
        for line in config_lines_reverse:
            if line.startswith("interface"):
                # If any of the interfaces listed in permanent_interfaces is a substring of the line
                if any(interface in line for interface in self.destination_device.PHYSICAL_INTERFACES_LIST):
                    self.unload_interface(line, wrap_command=False)
                else:
                    self.write_command(f"no {line}")

        # Get all the lines until the first interface
        interface_index = next((i for i, line in enumerate(config_lines) if line.startswith("interface")), None)
        config_lines_until_interfaces = config_lines[:interface_index]
        # Get all the lines that are not preceded with space -> assumes that they are unloaded as part of the main line unload
        main_lines = [line for line in config_lines_until_interfaces if (not line.startswith(' ') and not "exit" in line)]
        for line in main_lines[:1:-1]:  # Traverse from bottom to top again
            if "license activate" in line:
                continue
            # NOTE: Ignore cases that the "no" prefix will not work, expect the user to manually unload these in the loop below
            self.write_command(f"no {line}")

        # Finally, if user knows that there are configuration leftovers, unload it manually
        if unload_specific_commands is not None:
            for command in unload_specific_commands:
                self.write_command(command)

        self.write_command("hostname localhost")
        self.write_command("end")
        self.flush()

        # NOTE: By default, keep check to False because "show running-config" takes ~4s to return response
        if check_is_empty and not self.is_config_empty(self.write_command("show running-config")):
            logger.error(f"Config not fully unloaded for device {self.destination_device.hostname}")
            return
        logger.info(f"Config unloading effort finished for device {self.destination_device.hostname}")

    def is_config_empty(self, configuration: str, except_lines: Optional[List[str]]=None) -> bool:
        """
        Checks if the configuration of the device is fully empty and return boolean.
        """
        config_lines = configuration.split("\n")
        if "show running-config" not in config_lines[0] or "localhost#" not in config_lines[-1]:
            logger.debug(f"Returned config is not okay: {config_lines}")
            return False

        # Remove lines that should not be checked (lines in `except_lines` list)
        config_lines = [line for line in config_lines if all(exception not in line for exception in except_lines)]

        # Ensure empty interfaces pattern
        interface_lines = config_lines[1:-1]
        for i in range(len(interface_lines)):
            if i % 2 == 0:
                line = interface_lines[i].split()
                if line[0] != "interface" or \
                   line[1] not in self.destination_device.PHYSICAL_INTERFACES_LIST:
                    return False
            else:
                if "exit" not in interface_lines[i]:
                    return False
        return True

    @check_connection
    def ping(self, ip: str, nbr_packets: int = 1, ping_timeout: int = 1) -> str:
        """
        Sends a ping command to a specified IP address from the device.
        Supports both Linux and OneOS devices.
        """
        if self.destination_device.type == "oneos":
            response = self.write_command(f"ping {ip} -n {nbr_packets} -w {ping_timeout}")
            logger.info(f"Ping {nbr_packets * 5} packets at IP: {ip}")
            return response
        elif self.destination_device.type == "linux":
            response = self.write_command(f"ping {ip} -c {nbr_packets} -W {ping_timeout}")
            logger.info(f"Ping {nbr_packets} packets at IP: {ip}")
            return response
        else:
            raise NotImplementedError(f"Ping not implemented for device type {self.destination_device.type}")

    @check_device_type("linux")
    def hping3(
        self,
        destination_ip: str,
        nbr_packets: Optional[int] = None,
        interval: Optional[str] = None,
        flood: bool = False,
        port: Optional[int] = None,
        type: Optional[str] = None,
    ) -> None:
        """
        Execute hping3 command on the Linux device.
        For more information about hping3, see https://linux.die.net/man/8/hping3
        """
        valid_types = ["tcp", "udp", "icmp", "rawip", "syn", "ack", "fin", "rst"]
        full_command = "hping3 "
        if nbr_packets is not None:
            full_command += f"-c {nbr_packets} "
        if interval is not None:
            full_command += f"-i {interval} "
        if flood:
            full_command += "--flood "
        if port is not None:
            full_command += f"-p {port} "
        if type is not None and type.lower() in valid_types:
            full_command += f"--{type} "
        self.write_command(full_command + destination_ip)

    @check_device_type("oneos")
    def reconfigure(self, commands_list: List[str]) -> None:
        """
        Reconfigures a OneOS device with a list of commands.
        The list of commands is expected to include the exact commands
            to be sent to the device, with their "exit" commands.

        Args:
            commands_list (List[str]): The list of commands to send to the device, excluding the "config terminal" and "end" commands.

        Raises:
            ValueError: If the device is not a OneOS device.
            ConnectionError: If the device is not connected.
        """
        logger.debug("Reconfiguring device ...")
        self.write_command("term len 0")
        self.write_command("config terminal")
        for command in commands_list:
            self.write_command(command)
        self.write_command("end")
        self.flush()
        logger.debug(f"reconfig commands: {' | '.join(commands_list)}")
        logger.info("Device reconfigured")


class TelnetConnection(Connection):
    """
    Represents a Telnet connection, always originating from the Host device to another device.
    It uses the telnetlib library to establish and manage the connection.
    """

    def __init__(self, timeout: int=10):
        super().__init__(timeout)
        self.resulting_telnet_connection = telnetlib.Telnet()  # Not connected

    @Connection.check_occupied
    def connect(self, destination_device: "Device", destination_ip: str) -> Connection:
        """
        First connection from Host Device to any other Device.
        It uses an instantiated telnetlib.Telnet object, which is not connected yet.
        Returns the resulting Connection object, which carries the now connected telnetlib.Telnet object.

        Args:
            destination_device (Device): The device object representing the destination device.
                                         This object should contain the necessary credentials.
            destination_ip (str): The IP address of the destination device.

        Returns:
            Connection: The resulting Connection object, which carries the now connected telnetlib.Telnet object.

        Raises:
            ConnectionAbortedError: If the connection could not be established.
        """
        # Important to have a default, because if device is being reused (connect, set sudo, disconnect and then connect again),
        # it will connect as a user but the prompt_symbol will be #, which is not the default for a user
        self.prompt_symbol = destination_device.DEFAULT_PROMPT_SYMBOL
        self.destination_device = destination_device
        self.destination_ip = destination_ip

        self.resulting_telnet_connection.open(host=self.destination_ip, timeout=self.timeout)
        possible_login_prompts = [b"Username:", b"login:"]
        self._write_credentials(possible_login_prompts, destination_device.username)
        possible_password_prompts = [b"Password:"]
        self._write_credentials(possible_password_prompts, destination_device.password)
        encoded_prompt = self.prompt_symbol.encode("ascii")
        self.resulting_telnet_connection.read_until(encoded_prompt, self.timeout)

        if not self.is_connected:
            raise ConnectionAbortedError("Connection aborted: Could not connect")
        logger.info(f"Connected to {self.destination_device.hostname} at {self.destination_ip}")
        return self

    def _write_credentials(self, list_str_to_expect: List[str], str_to_write: str) -> None:
        if self.resulting_telnet_connection is not None:
            n, match, previous_text = self.resulting_telnet_connection.expect(
                list_str_to_expect, self.timeout
            )
            if n != -1:
                self.resulting_telnet_connection.write(str_to_write.encode("ascii") + b"\r")
            else:
                logging.error(f"EOFError: No match found. Match: {match}, Previous text: {previous_text}")
                raise EOFError
        else:
            logging.error("No connection object from telnetlib found. It has been closed or was never created.")

    @Connection.check_occupied
    def disconnect(self) -> None:
        self.resulting_telnet_connection.close()
        if self.is_connected:
            raise ConnectionError("Connection could not be closed")
        logger.info(f"Disconnected from {self.destination_device.hostname} at {self.destination_ip}")

    @property
    def is_connected(self) -> bool:
        """
        Checks if the telnet connection is active, by attempting to access the socket.
        """
        if not self.resulting_telnet_connection:
            return False
        try:
            # If the Telnet connection is not active, this will raise an exception
            _ = self.resulting_telnet_connection.get_socket().getsockopt(socket.SOL_SOCKET, socket.SO_TYPE)
            return True
        except Exception:
            return False


class TelnetCLIConnection(Connection):
    """
    Represents a CLI (Command Line Interface) connection over Telnet, used as a hop from a connected device to another.
    It is the equivalent of an open terminal, and then the developer executing "telnet <ip>".
    For the initialization of this connection type, a already connected TelnetConnection object is required.

    When instantiated properly, the base connection is set to "occupied" and is not available for use by other connections.
    If this TelnetCLIConnection object is used as a base connection for another TelnetCLIConnection object, then it is also set as "occupied".
    The base connections are freed when the exit() or the disconnect() methods of this object are called.
    """

    def __init__(self, source_connection: "TelnetConnection", timeout: int=10):
        super().__init__(timeout)
        self.source_connection = source_connection
        if self.source_connection.resulting_telnet_connection is None:
            raise ConnectionError("No connection object found during TelnetCLIConnection instantiation.")
        if self.source_connection._is_occupied:
            raise ConnectionRefusedError("The source connection is already in use. Please close the connections that use it first.")
        self.resulting_telnet_connection = self.source_connection.resulting_telnet_connection
        self._is_disconnected = True  # For internal use, to monitor when explicitly disconnecting. is_connected checks for the socket only

    @Connection.check_occupied
    def connect(self, destination_device: "Device", destination_ip: str) -> Connection:
        """
        This method uses the source device's connection to establish a new Telnet connection to the next destination device.

        Args:
            destination_device (Device): The device object representing the destination device.
                                         This object should contain the necessary credentials.
            destination_ip (str): The IP address of the destination device.

        Returns:
            Connection: The resulting Connection object, which carries the now connected telnetlib.Telnet object.

        Raises:
            ConnectionRefusedError: If the connection is refused by the destination device.
            ConnectionAbortedError: If the necessary prompts are not retrieved during the login process.
            ConnectionError: If the connection could not be established or if there is an error during the connection process.
        """
        # Will need the info from the source_connection's destination_device until fully connecting (i.e. prompt_symbol)
        self.destination_device = self.source_connection.destination_device  # pass by reference
        self.prompt_symbol = self.source_connection.prompt_symbol
        username = destination_device.username
        password = destination_device.password

        response = self.write_command(f"telnet {destination_ip}", expected_prompt_pattern=[b'Username:', b'login:'], timeout=self.timeout)
        if "Connection closed by foreign host" in response:
            raise ConnectionRefusedError("Connection refused: could not connect to next hop.")
        if "Username:" not in response and "login:" not in response:
            raise ConnectionAbortedError("Connection aborted: Username or Login prompts not retrieved.")
        response = self.write_command(username + '\n', expected_prompt_pattern=[b'Password:'], timeout=self.timeout)
        if "Password:" not in response:
            raise ConnectionAbortedError("Connection aborted: Password prompt not retrieved.")
        response = self.write_command(password + '\n', expected_prompt_pattern=[b'connected', b'Welcome'], timeout=self.timeout)
        if "connected" not in response and "Welcome" not in response:
            raise ConnectionError("Connection error: Could not connect to next hop.")

        # Finally, update with the most recent connection info
        self.source_connection._is_occupied = True
        self.destination_device = destination_device
        self.prompt_symbol = self.destination_device.DEFAULT_PROMPT_SYMBOL
        self.destination_ip = destination_ip
        self._is_disconnected = False

        if not self.is_connected:
            raise ConnectionError("Connection could not be established")
        logger.info(f"Connected " \
                  + f"from {self.source_connection.destination_device.hostname} " \
                  + f"to {self.destination_device.hostname} at {self.destination_ip}")
        return self

    @property
    def is_occupied(self) -> bool:
        """
        Checks if the connection is currently in use.
        """
        if not self.is_connected:
            self.source_connection._is_occupied = False
            self._is_occupied = False
        return self._is_occupied

    @Connection.check_occupied
    def disconnect(self) -> None:
        self.exit()

    @Connection.check_occupied
    def exit(self) -> Union["TelnetConnection", "TelnetCLIConnection"]:
        """
        Exits the current connection, but it doesn't close the socket, just returns to the previous hop.

        Returns:
            Union[TelnetConnection, TelnetCLIConnection]: The previous hop's connection object.
        """
        self.source_connection._is_occupied = False
        # Write "exit" to jump back to previous hop
        self.write_command(command="exit",
                           expected_prompt_pattern= [
                               b'closed', b'Connection closed by foreign host.'
                            ],
                           timeout=self.timeout)
        # Obligatory to return connection object because it might be of a different type
        self._is_disconnected = True
        logger.info(f"Jumped back to previous hop at device {self.source_connection.destination_device.hostname}")
        return self.source_connection

    @property
    def is_connected(self) -> bool:
        return self.source_connection.is_connected and not self._is_disconnected



#!/usr/bin/env python3

"""
File: test_ipsec.py (OLD VERSION)

This file contains tests for the IPSEC functionality.
Each test function in this module tests a specific aspect of the IPSEC functionality.
The name of the test function should clearly indicate what aspect it's testing.

Setup:
    For the setup, see `./ipscec_controlplane/test/scripts/config_ipsec/README.md`

Requirements:
    Python version should be minimum 3.7
    See requirements.txt for required python packages

Test Overview:
    For a test overview, see `./ipsec_controlplane/test/scripts/README.md`

FIXME: setup9, setup19 (Oragne Setups) not working
FIXME: setup27 not working, not able to ping the IPv6 addresses
"""

import os
import re
import sys
import json
import logging
import socket
from time import sleep
from typing import Optional, Dict, List, Tuple, Set

import pytest

# Add the root directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))
import src.static_utils
from src.device import OneOS6Device, RADIUSServer
from src.connection import TelnetConnection
from conftest import RED, GREEN, YELLOW, NC, OK, NOK, SKIPPED  # Colouring
from conftest import ROOT_PATH, IPSEC_CFG_DIR_NAME, IPSEC_JSON_NAME
from conftest import SHOW_CRYPTO_DIR, BSA_DIR, RADIUS_CFG_DIR_NAME  # Paths and file names
from conftest import TEST_SETUPS_GENERIC, TEST_SETUPS_ALGORITHMS, json_config  # Info from JSON


logger = logging.getLogger(__name__)


@pytest.mark.ipsec
@pytest.mark.generic
@pytest.mark.flaky(reruns=2)
@pytest.mark.parametrize("setup_marker, description, keywords", TEST_SETUPS_GENERIC)
def test_ipsec_generic(setup_marker: str, description: str, keywords: List[str], sudo_password) -> None:
    src.static_utils.print_banner(f"GENERIC - {setup_marker.upper()}", description)
    setup_info = (test_ipsec_generic.__name__, setup_marker)
    local_radius_connection_used = None

    intf_info = setup_interfaces(sudo_password)

    if "radius" in keywords:
        local_radius_connection_used = setup_radius(setup_info)

    connection_vm_a, connection_vm_b, connection_vm_c, _ = setup_connections(setup_info)

    try:
        logger.debug("Start performing checks ...")

        showCryptoIpsecSA_referenceData_VM_A = get_reference_data(setup_info, "VM_A", "show_crypto_ipsec_sa")
        showCryptoIpsecSA_referenceData_VM_B = get_reference_data(setup_info, "VM_B", "show_crypto_ipsec_sa")

        assert_nbr_active_sa(
            (connection_vm_a, showCryptoIpsecSA_referenceData_VM_A),
            (connection_vm_b, showCryptoIpsecSA_referenceData_VM_B),
        )

        assert_show_crypto_ipsec_sa(
            connection_vm_a, showCryptoIpsecSA_referenceData_VM_A,
            data_remove_lines_containing = {"spi", "sec)", "#", "protected vrf: default", "protected vrf: (none)"},
            expected_data_remove_lines_containing = {"spi", "sec)", "#"},
        )
        assert_show_crypto_ipsec_sa(
            connection_vm_b, showCryptoIpsecSA_referenceData_VM_B,
            data_remove_lines_containing = {"spi", "sec)", "#", "protected vrf: default", "protected vrf: (none)"},
           expected_data_remove_lines_containing = {"spi", "sec)", "#"},
        )

        assert_show_crypto_acl(connection_vm_a, get_reference_data(setup_info, "VM_A", "show_crypto_acl"))
        assert_show_crypto_acl(connection_vm_b, get_reference_data(setup_info, "VM_B", "show_crypto_acl"))

        assert_data_test(*get_script_data(setup_info), setup_info, connection_vm_a, connection_vm_b)

        assert_show_policy_interface(connection_vm_a, get_reference_data(setup_info, "VM_A", "show_policy_interface"))

        assert_show_commands_dont_fail(connection_vm_a)
        assert_show_commands_dont_fail(connection_vm_b)

        logger.info(f"{GREEN}{OK}{description} - TEST PASSED{NC}")

    finally:
        cleanup_connections(connection_vm_a, connection_vm_b, connection_vm_c)
        cleanup_interfaces(sudo_password, intf_info)
        if local_radius_connection_used:
            cleanup_radius(local_radius_connection_used)


@pytest.mark.ipsec
@pytest.mark.algorithms
@pytest.mark.flaky(reruns=2)
@pytest.mark.parametrize("setup_marker, description, keywords", TEST_SETUPS_ALGORITHMS)
def test_ipsec_algorithms(setup_marker: str, description: str, keywords: List[str], sudo_password) -> None:
    src.static_utils.print_banner(f"ALGORITHMS - {setup_marker.upper()}", description)
    setup_info = (test_ipsec_algorithms.__name__, setup_marker)

    intf_info = setup_interfaces(sudo_password)
    connection_vm_a, connection_vm_b, _, _ = setup_connections(setup_info)

    try:
        logger.debug("Start performing checks ...")

        showCryptoIpsecSA_referenceData_VM_A = get_reference_data(setup_info, "VM_A", "show_crypto_ipsec_sa")
        showCryptoIpsecSA_referenceData_VM_B = get_reference_data(setup_info, "VM_B", "show_crypto_ipsec_sa")

        assert_nbr_active_sa(
            (connection_vm_a, showCryptoIpsecSA_referenceData_VM_A),
            (connection_vm_b, showCryptoIpsecSA_referenceData_VM_B),
        )

        assert_algorithms(
            connection_vm_a,
            get_reference_data(setup_info, "VM_A", "show_crypto_ikev2_sa_detailed")
        )
        assert_algorithms(
            connection_vm_b,
            get_reference_data(setup_info, "VM_B", "show_crypto_ikev2_sa_detailed")
        )

        logger.info(f"{GREEN}{OK}{description} - TEST PASSED{NC}")
    finally:
        cleanup_connections(connection_vm_a, connection_vm_b)
        cleanup_interfaces(sudo_password, intf_info)


############################################################################### Aux functions


def check_radius_connection(radius_preference: str) -> None:
    response = _get_radius_connection_ok(radius_preference)
    if response == "OK":
        logger.info(f"{GREEN}{OK}RADIUS connection is OK{NC}")
    else:
        if response == "RADIUS_NOT_SET":
            logger.error(f"{RED}{NOK}RADIUS is required but is not properly set up{NC}")
        elif response == "RADIUS_NOT_RESPONDING":
            logger.error(f"{RED}{NOK}RADIUS is not responding{NC}")
        elif response == "RADIUS_IP_NOT_VALID":
            logger.error(f"{RED}{NOK}RADIUS IP is not valid{NC}")
        else:
            logger.error(f"{RED}{NOK}Unknown error while checking RADIUS Server{NC}")
        logger.warning(f"{YELLOW}{SKIPPED}Test fails because of RADIUS, not IPSEC{NC}")  # ?
        pytest.skip("RADIUS test is skipped")


def _get_radius_connection_ok(radius_preference: str) -> str:
    if radius_preference not in ["local", "remote"]:
        logger.critical(f"{RED}{NOK}Invalid radius_preference, expected 'local' or 'remote' but got {radius_preference} instead. Check JSON file{NC}")
        raise ValueError
    status = _check_radius_settings_in_json(radius_preference)
    if status != "OK":
        return status

    radius_ip = json_config["RADIUS"][radius_preference]["radius_ip"]
    if not src.static_utils.is_valid_ip(radius_ip):
        logger.critical(f"{RED}RADIUS IP: {radius_ip}{NC}")
        return "RADIUS_IP_NOT_VALID"

    ping_packet_loss = src.static_utils.get_packet_loss(src.static_utils.ping(radius_ip, count=2))
    if ping_packet_loss == "100" or ping_packet_loss is None or ping_packet_loss == "":
        return "RADIUS_NOT_RESPONDING"
    return "OK"


def _check_radius_settings_in_json(radius_preference: str) -> None:
    try:
        if json_config["RADIUS"][radius_preference]["radius_ip"]:
            return "OK"
        else:
            raise KeyError
    except KeyError as e:
        logger.error(f"{RED}Key '{str(e)}' does not exist in the JSON.{NC}")
    return "RADIUS_NOT_SET"


def assert_algorithms(connection: "TelnetConnectionIPSEC", ref_data: str) -> None:
    """
    Example line: "Encr: AES-CBC, keysize: 256, Hash: SHA512, DH Grp: 16"
    """
    encr_info: dict = _parse_encr_info(re.findall(r"(Encr: .*)", ref_data)[0])
    assert_config_is_matching(connection, encr_info)
    assert_dpd(connection, ref_data)


def assert_config_is_matching(connection: "TelnetConnectionIPSEC", encr_info: Dict[str, str]) -> None:
    """
    Checks the current configuration of the connection and compares it with the expected encryption information.

    Parameters:
        connection (TelnetConnectionIPSEC): The connection to check.
        encr_info (Dict[str, str]): A dictionary containing the expected encryption information.

    Raises:
        AssertionError: If the current configuration does not match the expected configuration,
                        or if the completion of any encryption, integrity, or group is wrong.
    """
    try:
        # Calls "show running-config" (5s overhead)
        crypto_ikev2_proposal_name = assert_current_config(connection, encr_info)
    except AssertionError as err:
        logger.error(f"{RED}{NOK}Current config is not as expected - {connection.destination_device.hostname}{NC}")
    # Check encryption, integrity and group completions
    for key, value in encr_info.items():
        try:
            assert_completion(connection, key, value, crypto_ikev2_proposal_name)
        except AssertionError as err:
            logger.error(f"{RED}{NOK}Completion of '{key} ?' is wrong - {connection.destination_device.hostname}{NC}")
            raise err
    logger.info(f"{GREEN}{OK}Config is matching - {connection.destination_device.hostname}{NC}")


def assert_current_config(connection: "TelnetConnectionIPSEC", encr_info: str) -> str:
    current_config = connection.write_command("show running-config")
    subconfig_split, crypto_ikev2_proposal_name = _get_crypto_ikev2_proposal_subconfig(current_config)
    # Check if the config matches the reference data
    count_checked = 0
    for line in subconfig_split:
        if "encryption" in line:
            assert encr_info["encryption"] in line, f'Expected "{encr_info["encryption"]}" in line "{line}"'
            count_checked += 1
            continue
        if "integrity" in line:
            assert encr_info["integrity"] in line, f'Expected "{encr_info["integrity"]}" in line "{line}"'
            count_checked += 1
            continue
        if "group" in line:
            assert encr_info["group"] in line, f'Expected "{encr_info["group"]}" in line "{line}"'
            count_checked += 1
            continue
    if count_checked != 3:
        raise AssertionError("Did not perform exactly 3 checks, for encryption, integrity and group. Check the config:\n" + '\n'.join(subconfig_split))
    return crypto_ikev2_proposal_name


def assert_completion(connection: "TelnetConnectionIPSEC", key: str, value: str, name: str) -> None:
    response = _get_crypto_ikev2_proposal_help(connection, name, key)
    if response is None:
        raise AssertionError(f'"{key} ?" did not return a proper response: {response}')
    response_split = response.split('\n')
    if len(response_split) < 3:
        raise AssertionError(f'"{key} ?" did not return a proper response: {response_split}')
    response_split = response_split[3:-1]  # Discard header and footer

    found_value = False
    for line in response_split:
        if value in line:
            found_value = True
            break
    assert found_value, f'"{value}" not found in the list of possible completions:' + '\n'.join(response_split)


def assert_dpd(connection: "TelnetConnectionIPSEC", ref_data: str, checks_iterations: int=3, checks_sleep: int=10) -> None:
    """
    Checks the DPD configuration and the increment of local and remote message IDs over a number of iterations.
    A DPD request is sent every 10 seconds, so the message IDs should increment at least once every 10 seconds.

    Args:
        connection (TelnetConnectionIPSEC): The connection to check.
        ref_data (str): The reference data to compare against.
        checks_iterations (int, optional): The number of times to check the message IDs. Defaults to 3.
        checks_sleep (int, optional): The number of seconds to sleep between checks. Defaults to 10.

    Raises:
        AssertionError: If the DPD configuration does not match the expected configuration,
                        or if the message IDs do not increment as expected.
    """
    response = connection.write_command("show crypto ikev2 sa detailed")

    dpd_pattern = re.compile(r"DPD configured (.*)")
    dpd_configuration = re.search(dpd_pattern, response).group(1).strip()
    dpd_configuration_expected = re.search(dpd_pattern, ref_data).group(1).strip()
    assert dpd_configuration == dpd_configuration_expected, \
        f"Expected {dpd_configuration_expected}, but got {dpd_configuration} instead"

    local_req_msg_id_pattern  = re.compile(r"Local req msg id:  (\d+).*")
    remote_reg_msg_id_pattern = re.compile(r"Remote reg msg id:  (\d+).*")

    initial_local_req_msg_id  = int(local_req_msg_id_pattern.search(response).group(1))
    initial_remote_reg_msg_id = int(remote_reg_msg_id_pattern.search(response).group(1))
    for _ in range(checks_iterations):
        logger.debug("DPD Check: Iteration %s - Waiting for counters to increment...", _)
        sleep(checks_sleep)
        response = connection.write_command("show crypto ikev2 sa detailed")
        local_req_msg_id  = int(local_req_msg_id_pattern.search(response).group(1))
        remote_reg_msg_id = int(remote_reg_msg_id_pattern.search(response).group(1))
        if local_req_msg_id > initial_local_req_msg_id and remote_reg_msg_id > initial_remote_reg_msg_id:
            logger.debug("DPD Check: SUCCESS - Both counters incremented after %ss", checks_sleep)
            logger.debug("DPD Check: Initial Counters - Local: %s,\tRemote: %s", initial_local_req_msg_id, initial_remote_reg_msg_id)
            logger.debug("DPD Check:   Final Counters - Local: %s,\tRemote: %s", local_req_msg_id, remote_reg_msg_id)
            initial_local_req_msg_id = local_req_msg_id
            initial_remote_reg_msg_id = remote_reg_msg_id
        else:
            logger.error("DPD Check: Failure - Counters did not increment after %ss on device %s", checks_sleep, connection.destination_device.hostname)
    logger.info(f"{GREEN}{OK}DPD: Counters incremented after {checks_iterations} checks - {connection.destination_device.hostname}{NC}")


def assert_show_policy_interface(connection: "TelnetConnectionIPSEC", expected_data: Optional[str] = None) -> None:
    if expected_data is None:
        logger.info(f"{YELLOW}{SKIPPED}show policy-interface for device {connection.destination_device.hostname}{NC}")
        return
    data = connection.parse_show_policy_interface()
    reference_data = clean_lines(expected_data, from_id=1)
    _check_diff_in_files(
        data, reference_data, test_type="show policy interface",
        device_name=connection.destination_device.hostname
    )
    logger.info(f"{GREEN}{OK}show policy-interface - {connection.destination_device.hostname}{NC}")


def assert_data_test(nbr_packets_expected: int, test_commands: List[str], setup_info: Tuple[str, str], *connections: "TelnetConnectionIPSEC") -> None:
    if nbr_packets_expected is None and test_commands is None:
        for connection in connections:
            logger.info(f"{YELLOW}{SKIPPED}Data Test for device {connection.destination_device.hostname}{NC}")
        return

    logger.info("Starting Data Test...")
    # The first execution of the script is for setup purposes, results are not relevant.
    logger.debug("Sending traffic for the First Time")
    src.static_utils.execute_shell_commands_on_host(test_commands, quiet=True)
    sleep(3)
    for connection in connections:
        connection.clear_crypto_counters()
    sleep(1)
    logger.debug("Sending traffic for the Second Time")
    for connection in connections:
        connection.clear_policy_interface()

    assert_nbr_packets(get_successful_pings(test_commands), nbr_packets_expected)
    sleep(5)

    reference_data = {}
    for i, connection in enumerate(connections, start=1):
        vm_name = f"VM_{chr(64 + i)}"  # This will generate VM_A, VM_B, VM_C, etc.
        reference_data[vm_name] = get_reference_data(setup_info, vm_name, "show_crypto_ipsec_sa_data")

    remove_lines_containing = {"spi", "sec)", "protected vrf: default", "protected vrf: (none)"}
    for i, connection in enumerate(connections, start=1):
        vm_name = f"VM_{chr(64 + i)}"  # This will generate VM_A, VM_B, VM_C, etc.
        assert_show_crypto_ipsec_sa(connection, reference_data[vm_name], remove_lines_containing)
    logger.info(f"{GREEN}{OK}Data Test{NC}")


def assert_show_commands_dont_fail(connection: "TelnetConnectionIPSEC") -> None:
    assert connection.write_command("show crypto call admission"),            f'{RED}command "show crypto call admission" did not return anything{NC}'
    assert connection.write_command("show crypto call admission statistics"), f'{RED}command "show crypto call admission statistics" did not return anything{NC}'
    assert connection.write_command("show crypto ikev2 sa"),                  f'{RED}command "show crypto ikev2 sa" did not return anything{NC}'
    assert connection.write_command("show crypto ikev2 sa detailed"),         f'{RED}command "show crypto ikev2 sa detailed" did not return anything{NC}'
    assert connection.write_command("show crypto ikev2 session"),             f'{RED}command "show crypto ikev2 session" did not return anything{NC}'
    assert connection.write_command("show crypto ikev2 session detailed"),    f'{RED}command "show crypto ikev2 session detailed" did not return anything{NC}'
    assert connection.write_command("show crypto ikev2 stats"),               f'{RED}command "show crypto ikev2 stats" did not return anything{NC}'
    assert connection.write_command("show crypto ipsec client ezvpn"),        f'{RED}command "show crypto ipsec client ezvpn" did not return anything{NC}'
    assert connection.write_command("show crypto ipsec sa"),                  f'{RED}command "show crypto ipsec sa" did not return anything{NC}'
    assert connection.write_command("show crypto ipsec sa detail"),           f'{RED}command "show crypto ipsec sa detail" did not return anything{NC}'
    assert connection.write_command("show crypto ipsec sa identity"),         f'{RED}command "show crypto ipsec sa identity" did not return anything{NC}'
    assert connection.write_command("show crypto ipsec sa identity detail"),  f'{RED}command "show crypto ipsec sa identity detail" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp key"),                f'{RED}command "show crypto isakmp key" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp policy"),             f'{RED}command "show crypto isakmp policy" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp profile"),            f'{RED}command "show crypto isakmp profile" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp sa"),                 f'{RED}command "show crypto isakmp sa" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp sa count"),           f'{RED}command "show crypto isakmp sa count" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp sa detail"),          f'{RED}command "show crypto isakmp sa detail" did not return anything{NC}'
    assert connection.write_command("show crypto isakmp statistics"),         f'{RED}command "show crypto isakmp statistics" did not return anything{NC}'
    assert connection.write_command("show expert crypto config"),             f'{RED}command "show crypto config" did not return anything{NC}'
    assert connection.write_command("show expert crypto dns"),                f'{RED}command "show crypto dns" did not return anything{NC}'
    assert connection.write_command("show expert crypto cmid"),               f'{RED}command "show crypto cmid" did not return anything{NC}'
    assert connection.write_command("show crypto session"),                   f'{RED}command "show crypto session" did not return anything{NC}'
    assert connection.write_command("show crypto session detail"),            f'{RED}command "show crypto session detail" did not return anything{NC}'
    assert connection.write_command("show crypto gkm group"),                 f'{RED}command "show crypto gkm group" did not return anything{NC}'
    assert connection.write_command("show crypto gkm group name g-poc"),      f'{RED}command "show crypto gkm group name g-poc" did not return anything{NC}'
    logger.info(f"{GREEN}{OK}All other show commands returned data - {connection.destination_device.hostname}{NC}")


def assert_nbr_packets(nbr_packets: int, nbr_packets_expected: int) -> None:
    # Probably first arg will arrive as string. Don't allow implicit conversion
    assert nbr_packets == nbr_packets_expected or nbr_packets == nbr_packets_expected - 1, \
        f"Expected {nbr_packets_expected} packets, but got {nbr_packets} instead"


def assert_show_crypto_acl(connection: "TelnetConnectionIPSEC", expected_data: Optional[str] = None) -> None:
    if expected_data is None:
        logger.info(f"{YELLOW}{SKIPPED}show crypto acl for device {connection.destination_device.hostname}{NC}")
        return
    data = connection.parse_show_crypto_acl()
    reference_data = clean_lines(expected_data, from_id=2, to_id=-1)
    _check_diff_in_files(data, reference_data, test_type="show crypto acl",
                         device_name=connection.destination_device.hostname)
    logger.info(f"{GREEN}{OK}show crypto acl - {connection.destination_device.hostname}{NC}")


def assert_nbr_active_sa(*info: Tuple["TelnetConnectionIPSEC", str], timeout: int = 10) -> None:
    # Parse pairs
    connection_data = {}
    for connection, reference_data in info:
        if not connection.is_connected:
            raise ConnectionError(f"Connection to {connection.destination_device.hostname} is not established")
        expected_nbr = get_nbr_active_sessions_from_reference_file(reference_data)
        if expected_nbr == 0:
            logger.info(f"{YELLOW}{SKIPPED}no SA expected to be ACTIVE for device {connection.destination_device.hostname}{NC}")
            continue
        # List items will be used in the same order as they are passed
        connection_data[connection] = expected_nbr

    logger.info("Waiting for SAs")

    nbr_iterations = 1
    initial_timeout = timeout
    while True:
        if timeout <= 0:
            logger.error(f"{RED}{NOK}Waiting for SAs: Timeout reached{NC}")
            raise TimeoutError
        status_active_in_connections = []
        for connection, expected_nbr in connection_data.items():
            status_active_in_connections.append(
                connection.get_nbr_active_sa() == expected_nbr
            )

        # if any(status_active_in_connections):
        if all(status_active_in_connections):
            logger.info(f"{GREEN}{OK}nbr active SA{NC}")
            break

        if nbr_iterations == 5:
            for connection, _ in connection_data.items():
                connection.clear_crypto_sa()
        logger.info(f"Waiting for tunnel to appear ACTIVE ... {(nbr_iterations-1)*10}s/{initial_timeout*10}s")
        nbr_iterations += 1
        timeout -= 1
        sleep(10)
    sleep(2)
    return


def assert_show_crypto_ipsec_sa(
    connection: "TelnetConnectionIPSEC",
    expected_data: str,
    data_remove_lines_containing: Optional[Set] = None,
    expected_data_remove_lines_containing: Optional[Set] = None,
) -> None:
    data = connection.parse_show_crypto_ipsec_sa(remove_lines_containing=data_remove_lines_containing)
    reference_data = _parse_expected_data(expected_data, expected_data_remove_lines_containing)
    _check_diff_in_files(data, reference_data, test_type="show crypto ipsec sa",
                         device_name = connection.destination_device.hostname)
    logger.info(f"{GREEN}{OK}show crypto ipsec sa - {connection.destination_device.hostname}{NC}")


def clean_lines(data: str, from_id: Optional[int]=None, to_id: Optional[int]=None) -> List[str]:
    return [line for line in (l.strip() for l in data.split('\n')[from_id:to_id]) if line]


def get_script_data(setup_info: Tuple[str, str]) -> Tuple[int, List[str]]:
    test_name, setup_marker = setup_info
    data_file = json_config[test_name][setup_marker].get("data_file")
    if data_file is not None:

        data_file = os.path.join(ROOT_PATH, IPSEC_CFG_DIR_NAME, SHOW_CRYPTO_DIR, data_file)

        with open(data_file) as f:
            data = f.read().split("\n")
            nbr_packets_expected = int(data[1][1:])  # First character will always be a "#"
            test_commands = data[2:]
            test_commands = [line for line in test_commands if line]  # Clear empty lines
            return nbr_packets_expected, test_commands
    return None, None


def cleanup_connections(*connections: "TelnetConnectionIPSEC") -> None:
    for connection in connections:
        if connection is not None:
            cleanup_device(connection)


def cleanup_interfaces(password, interface_info: List[Dict[str, str]]) -> None:
    for interface in interface_info:
        try:
            intf_name = interface["name"]
            if "ip" in interface:
                intf_ip = interface["ip"]
                src.static_utils.del_interface_ip(intf_name, intf_ip, password)
            elif "ipv6" in interface:
                intf_ip = interface["ipv6"]
                src.static_utils.del_interface_ip(intf_name, intf_ip, password, 64)
            else:
                logger.error(f"No 'ip' or 'ipv6' key found in interface dictionary.")
                raise KeyError
        except ValueError:
            logger.error(f"{intf_ip} is not a valid IP address.")
        except KeyError:
            logger.error(f"Key not found in interface dictionary.")


def cleanup_radius(connection: TelnetConnection) -> None:
    connection.delete_interface_ip(
        interface_name=json_config["RADIUS"]["local"]["radius_interface"],
        ip_addr=json_config["RADIUS"]["local"]["radius_ip"],
    )


def get_successful_pings(commands: List[str]) -> int:
    response = src.static_utils.execute_shell_commands_on_host(commands, quiet=False)
    return sum("bytes from" in line for line in response.split('\n'))


def load_json(json_config_path: Optional[str] = None) -> Dict:
    try:
        if json_config_path is None:
            root_path = os.path.dirname(os.path.abspath(__file__))
            json_config_path = os.path.join(
                root_path, IPSEC_CFG_DIR_NAME, IPSEC_JSON_NAME
            )
        with open(json_config_path) as f:
            json_config = json.load(f)
        return json_config
    except FileNotFoundError:
        logger.error(f"File not found at path {json_config_path}")
        raise FileNotFoundError
    except json.JSONDecodeError:
        logger.error(f"File {json_config_path} is not a valid JSON file")
        raise json.JSONDecodeError


def cleanup_device(connection: "TelnetConnectionIPSEC", dont_disconnect=False) -> None:
    hostname = connection.destination_device.hostname
    connection.unload_config(
        # Collection of all the unload commands needed for all the setups
        # Added license key will NOT be deleted implicitly
        # While the "license activate acs" is unconfigured, key remains active until reboot
        unload_specific_commands=[
            "no crypto isakmp keepalive",  # Setup1
            "crypto engine disable",
            "no logging buffered debug",
            "no aaa authentication login aaaezvpn local",  # Setup2
            "no crypto map dynamic dynMap 10",
            "no crypto dynamic dynTemplate",
            "no crypto ipsec transform-set tsEzvpn",
            "no crypto isakmp profile ezvpnProf",
            "no ip local pool",
            "no ip access-list extended 110",
            "console timeout 600",  # Setup6
            "no crypto ikev2 dpd",  # Setup9
            "no ip access-list extended 100",  # Setup 11
            "no crypto call admission limit ipsec sa",
            "no crypto ikev2 fragmentation",  # Setup14
            "sp-fwd",  # Setup19
            "no crypto map local-address",  # Setup21
            "no crypto call admission limit ike in-nego",  # Setup 28
        ]
    )
    connection.clear_crypto_sa()
    answer = connection.has_open_sockets()
    assert not answer, f"{RED}{NOK}secd udp sessions not correctly closed - {hostname}{NC}"
    connection.disconnect() if not dont_disconnect else None


def get_reference_data(setup_info: Tuple[str, str], vm_name: str, show_command: str) -> Optional[str]:
    test_name, setup_marker = setup_info
    reference_file_name = json_config[test_name][setup_marker][vm_name].get(show_command)
    if reference_file_name is None:
        return None

    reference_file = os.path.join(ROOT_PATH, IPSEC_CFG_DIR_NAME, SHOW_CRYPTO_DIR, reference_file_name)
    with open(reference_file) as f:
        reference_data = f.read()
    return reference_data


def get_nbr_active_sessions_from_reference_file(reference_data: str) -> Optional[int]:
    reference_nbr_active_sa = 0
    for line in reference_data.split("\n"):
        line = line.strip()
        if line.startswith("Status:") and line.split(": ")[1] == "ACTIVE":
            reference_nbr_active_sa += 1
    return reference_nbr_active_sa


def ensure_clean_config_at_init(connection: "TelnetConnectionIPSEC", except_lines: Optional[str]=None) -> None:
    """
    If used, ~5s overhead because of "show running-config"
    """
    if except_lines is None:
        except_lines = ["license key add", "license activate"]

    connection.write_command("term len 0")
    if not connection.is_config_empty(connection.write_command("show running-config"), except_lines):
        logger.warning("Config is not empty at startup, trying to unconfigure")
        connection.flush_deep()
        cleanup_device(connection, dont_disconnect=True)
        if not connection.is_config_empty(connection.write_command("show running-config"), except_lines):
            raise ValueError("Config is not fully empty at startup, please manually erase config and reboot")
        else:
            logger.info("Config successfully unconfigured")


def check_and_activate_license(connection: "TelnetConnectionIPSEC", vm_name: str) -> None:
    if "license-key" in json_config[vm_name]:
        if not connection.is_license_sdwanprime_activated():
            if connection.is_license_sdwanprime_installed():
                connection.reconfigure([
                    "license activate sd-wan-prime",
                ])
            else:
                logger.critical("SD WAN license is not installed. Installing key from JSON and activating (might need reboot) ...")
                connection.reconfigure([
                    "license key add " + json_config[vm_name]["license-key"],
                    "license activate sd-wan-prime",
                ])


def setup_connection(setup_info: Tuple[str, str], vm_nbr: int, check_is_empty=True) -> "TelnetConnectionIPSEC":
    """
    check_is_empty: When developers kill the test, they leave config traces to the VMs' configs
                    With an overhead of ~5s, if this parameter is set to True, it ensures that the config is empty
    """
    if not 1 <= vm_nbr <= 4:
        raise ValueError("Invalid vm_nbr")
    vm_name = "VM_" + chr(64 + vm_nbr)

    # Skip cases where VM_C and VM_D are not expected. Assume that JSON doesn't provide config for them
    test_name, setup_marker = setup_info
    config_for_vm_doesnt_exist = (json_config[test_name][setup_marker].get(vm_name) is None)
    if config_for_vm_doesnt_exist:
        return None

    vm = OneOS6Device()
    cfg_dir_path = os.path.join(ROOT_PATH, IPSEC_CFG_DIR_NAME)
    try:
        connection = TelnetConnectionIPSEC().connect(vm, json_config[vm_name]["ip"])
    except socket.timeout as err:
        logger.error("Could not connect to device, caught socket.timeout.")
        logger.warning("Make sure the devices are running and reachable (check IPs in JSON file).")
        raise err

    if check_is_empty:
        ensure_clean_config_at_init(connection)

    connection.flush_deep()

    check_and_activate_license(connection, vm_name)

    connection.load_config(os.path.join(cfg_dir_path, BSA_DIR, json_config[test_name][setup_marker][vm_name]["config_file"]))

    # For Setup19: `_cont` files are patching up the initial config
    # Because of a race condition, some initial config is lost and must be repeated (trustpoints)
    patch = json_config[test_name][setup_marker][vm_name].get("config_file_patch")
    if patch is not None:
        connection.patch_config(os.path.join(cfg_dir_path, BSA_DIR, patch))

    # connection = src.static_utils.reboot_device(connection) if REBOOT_FLAG else connection
    return connection


def setup_connections(setup_info: Tuple[str, str]) -> Tuple[
    "TelnetConnectionIPSEC",            # VM_A
    "TelnetConnectionIPSEC",            # VM_B
    Optional["TelnetConnectionIPSEC"],  # VM_C
    Optional["TelnetConnectionIPSEC"],  # VM_D, not used so far
]:
    connection_a = setup_connection(setup_info, vm_nbr=1)
    connection_b = setup_connection(setup_info, vm_nbr=2)
    try:
        connection_c = setup_connection(setup_info, vm_nbr=3)
    except OSError:  # If config for VM_C exists but VM_C is not active, test should be skipped
        logger.error(f"{YELLOW}{SKIPPED}Found config, but could not connect to VM_C. Skipping test{NC}")
        connection_c = None
        pytest.skip("Could not connect on VM_C, skipping test")
    try:
        connection_d = setup_connection(setup_info, vm_nbr=4)
    except OSError:  # If config for VM_D exists but VM_D is not active
        logger.error(f"{YELLOW}{SKIPPED}Found config, but could not connect to VM_D. Skipping test{NC}")
        connection_d = None
        pytest.skip("Could not connect on VM_D, skipping test")
    return connection_a, connection_b, connection_c, connection_d


def setup_interfaces(password: str) -> List[Dict[str, str]]:
    interface_info = []
    for _, interface in json_config["HOST"]["interfaces"].items():
        ip = interface.get("ip", "")
        if ip:
            src.static_utils.set_interface_ip(interface["name"], interface["ip"], password)
            logger.info(f'IP {interface["ip"]} added to interface {interface["name"]}')
            interface_info.append({"name": interface["name"], "ip": interface["ip"]})

        ipv6 = interface.get("ipv6", "")
        if ipv6:
            src.static_utils.set_interface_ip(interface["name"], interface["ipv6"], password, netmask="64")
            logger.info(f'IP {interface["ipv6"]} added to interface {interface["name"]}')
            interface_info.append({"name": interface["name"], "ipv6": interface["ipv6"]})

        if not ip and not ipv6:
            logger.warning(f"No IP or IPv6 address provided for interface {interface['name']}")
    return interface_info


def setup_radius(setup_info: Tuple[str, str]) -> TelnetConnection:
    test_name, setup_marker = setup_info
    radius_preference = json_config[test_name][setup_marker]["radius_preference"]
    if radius_preference == "local":
        # set up ip addr in the specified interface of radius server
        connection_to_radius = TelnetConnection()
        radius_vm = RADIUSServer(
            username=json_config["RADIUS"]["local"]["username"],
            password=json_config["RADIUS"]["local"]["password"],
        )
        connection_to_radius.connect(radius_vm, json_config["RADIUS"]["local"]["ip"])

        if not is_radius_installed(connection_to_radius):
            logger.critical(f"{RED}{NOK}RADIUS is required but is not installed{NC}")
            raise ValueError
        if not is_radius_active(connection_to_radius):
            logger.critical(f"{RED}{NOK}RADIUS is installed but is not active{NC}")
            raise ValueError

        connection_to_radius.set_sudo(json_config["RADIUS"]["local"]["password"])

        # Move the config files to the RADIUS server
        user = json_config["RADIUS"]["local"]["username"]
        password = json_config["RADIUS"]["local"]["password"]
        src.static_utils.scp_file_to_home_dir(
            local_file_path  = os.path.join(ROOT_PATH, IPSEC_CFG_DIR_NAME, RADIUS_CFG_DIR_NAME, "authorize"),
            user_at_ip       = user + "@" + json_config["RADIUS"]["local"]["ip"],  # i.e. user@X.X.X.X
            password         = password,
        )
        connection_to_radius.write_command(f"mv /home/{user}/authorize /etc/freeradius/3.0/mods-config/files/authorize")
        src.static_utils.scp_file_to_home_dir(
            local_file_path  = os.path.join(ROOT_PATH, IPSEC_CFG_DIR_NAME, RADIUS_CFG_DIR_NAME, "clients.conf"),
            user_at_ip       = user + "@" + json_config["RADIUS"]["local"]["ip"],
            password         = password,
        )
        connection_to_radius.write_command(f"mv /home/{user}/clients.conf /etc/freeradius/3.0/clients.conf")

        # Might need to restart the RADIUS server here

        connection_to_radius.set_interface_ip(
            interface_name=json_config["RADIUS"]["local"]["radius_interface"],
            ip_addr=json_config["RADIUS"]["local"]["radius_ip"],
        )
    check_radius_connection(radius_preference)
    return connection_to_radius if radius_preference == "local" else None


def is_radius_installed(connection: TelnetConnection) -> bool:
    response = connection.write_command("dpkg -l | grep freeradius")
    return len(response.split('\n')[1:-1]) >= 5  # generic, common, config, utils, lib


def is_radius_active(connection: TelnetConnection) -> bool:
    connected_to_internet = False
    for _ in range (3):
        if src.static_utils.get_packet_loss(connection.ping("8.8.8.8", nbr_packets=3)) == "0":
            connected_to_internet = True
            break
    if not connected_to_internet:
        logger.critical(f"{RED}{NOK}RADIUS server is not connected to the internet{NC}")
        raise ConnectionError
    response = connection.write_command("ps aux | grep freeradius")
    return re.search(r"freerad ", response) is not None


def _get_crypto_ikev2_proposal_help(connection: "TelnetConnectionIPSEC", name: str, command: str) -> str:
    if name is None:
        raise ValueError("crypto ikev2 proposal <name> is not defined")
    previous_prompt_symbol = connection.prompt_symbol
    connection.prompt_symbol = '#'  # During (re)configuring, cannot search for i.e. "VM_A#", so focus only on the symbol "#"
    connection.write_command("configure terminal")
    connection.write_command("crypto ikev2 proposal " + name)  # Name does not even have to be correct
    response = connection.write_command(f"{command} ?")
    logger.debug(f'Checking possible completiong for "{command}":')
    logger.debug(response)
    connection.write_command("end")
    connection.prompt_symbol = previous_prompt_symbol
    return response


def _get_crypto_ikev2_proposal_subconfig(config: str) -> Tuple[List[str], str]:
    found = False
    lines_of_interest = []
    for line in config.split("\n"):
        if "crypto ikev2 proposal" in line:
            found = True
            crypto_ikev2_proposal_name = line.split()[3]
            lines_of_interest.append(line)
            continue
        if found:
            lines_of_interest.append(line)
        if found and "exit" in line:
            break
    if not found:
        raise AssertionError('No "crypto ikev2 proposal" found in the config')
    logger.debug("Subconfig collected: " + '\n'.join(lines_of_interest))
    return lines_of_interest, crypto_ikev2_proposal_name


def _parse_encr_info(encr_info: str) -> Dict[str, str]:
    """
    Comes in as:
        Encr: AES-CBC, keysize: 256, PRF: SHA512, Hash: SHA512, DH Grp: 16, Auth sign: PSK, Auth verify: PSK
    Is converted to:
        {"encryption": "aes-cbc-256", "integrity": "sha512", "group": "16"}
    """
    encryption = re.search(r'Encr: (\w+-\w+), keysize: (\d+)', encr_info)
    integrity = re.search(r'Hash: (\w+)', encr_info)
    group = re.search(r'DH Grp: (\d+)', encr_info)

    return {
        'encryption': f"{encryption.group(1).lower()}-{encryption.group(2)}" if encryption else None,
        'integrity': integrity.group(1).lower() if integrity else None,
        'group': group.group(1) if group else None
    }


def _check_diff_in_files(data: List[str], reference_data: List[str], test_type: str, device_name: str = None) -> None:
    data = _clean_data(data)
    reference_data = _clean_data(reference_data)

    # Find and store the differences between the two lists
    wrongly_retrieved_data = [item for item in data if item not in reference_data]
    missing_expected_data = [item for item in reference_data if item not in data]

    print_name = f" - {device_name}" if device_name else ""
    if wrongly_retrieved_data or missing_expected_data:
        logger.error(f"{RED}[NOK] - {test_type}{print_name}{NC}")
        logger.error(
            f"         Found {max(len(wrongly_retrieved_data), len(missing_expected_data))} "
            + "differences between the retrieved data and the expected data"
        )
        for line in wrongly_retrieved_data:
            logger.error(f'         Wrongly retrieved: "{line}"')
        for line in missing_expected_data:
            logger.error(f'         Missing expected: "{line}"')
        # Will only be printed if error is raised and logging level is DEBUG (most verbose)
        logger.debug("\n********\n" + "Retrieved data: " + '\n'.join(data))
        logger.debug("********\n" + "Reference data: " + '\n'.join(reference_data))
        raise AssertionError


def _parse_expected_data(expected_data: str, remove_lines_containing: Set[str] = None) -> List[str]:
    # Strip symbols and empty lines
    reference_data = [line.strip() for line in expected_data.split("\n") if line.strip()]
    # Remove lines containing certain strings, if remove_lines_containing is provided
    if remove_lines_containing:
        reference_data = [line for line in reference_data if not any(substring in line for substring in remove_lines_containing)]
    return sorted(reference_data)


def _clean_data(data) -> List[str]:
    cleaned_data = []
    for line in data:
        items = line.split()
        len_items = len(items)
        for i in range(len_items):
            items[i] = items[i].strip()
        cleaned_data.append(" ".join(items))
    return cleaned_data


class TelnetConnectionIPSEC(TelnetConnection):
    # Call constructor of base class
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # Decorator
    def get_response(command):
        """
        Decorator for getting the response of a command
        If response is provided, use it, otherwise write specified command
        """
        def decorator(func):
            def wrapper(self, *args, **kwargs):
                response = kwargs.pop("response", None)
                if response is None:
                    response = self.write_command(command)
                kwargs["response"] = response
                return func(self, *args, **kwargs)
            return wrapper
        return decorator

    @property
    def is_crypto_engine_enabled(self) -> bool:
        response = self.write_command("show crypto engine")
        if "disabled" in response:
            return False
        elif "enabled" in response:
            return True
        else:
            raise ValueError('Unexpected response from "show crypto engine"')

    @property
    def is_tunnel_active(self) -> bool:
        response = self.write_command("show crypto ipsec sa")
        if "Status: ACTIVE" in response:
            return True
        else:  # Includes "Status: INIT", "DELETE", "IDLE"
            return False

    @get_response("show udp sessions")
    def has_open_sockets(self, response: str) -> bool:
        secd4_sessions = "*:500" in response
        nat_sessions = "*:4500" in response
        secd6_sessions = "[::]:500" in response
        return (secd4_sessions or nat_sessions or secd6_sessions)


    @get_response("show crypto ipsec sa")
    def get_nbr_active_sa(self, response: Optional[str] = None) -> int:
        nbr_active_sa = 0
        for line in response.split("\n"):
            line = line.strip()
            if line.startswith("Status:") and line.split(": ")[1] == "ACTIVE":
                nbr_active_sa += 1
        return nbr_active_sa

    @get_response("show crypto ipsec sa")
    def parse_show_crypto_ipsec_sa(self, response: Optional[str] = None, remove_lines_containing: Optional[Set[str]] = None) -> List[str]:
        """
        Parses the output of `show crypto ipsec sa` and returns a list of strings.

        Each string is a sorted line of the output, with the following modifications:

            - Lines containing "spi", "sec)", "#", "protected vrf: default", or "protected vrf: (none)" are removed.
            - Occurrences of the address from DHCP "192.168.X.X" are masked.
            - The lines are sorted alphanumerically.

        Args:
            response (str, optional): The output of `show crypto ipsec sa`. If None, the function will get the response itself.

        Returns:
            List[str]: A list of strings representing the parsed and modified lines of the output.
        """
        if remove_lines_containing is None:
            remove_lines_containing = {
                "spi",
                "sec)",
                "#",
                "protected vrf: default",
                "protected vrf: (none)",
            }
        ip_pattern = re.compile(r"192\.168\.\d{1,3}\.\d{1,3}")
        tunnel_pattern = re.compile(r'Virtual-IpsecTunnel \d*')


        lines = [line for line in (l.strip() for l in response.split("\n")) if line]

        # Replace occurrences of IPs and tunnel numbers
        masked_lines = [tunnel_pattern.sub('Virtual-IpsecTunnel XX', ip_pattern.sub("192.168.XX.XX", line)) for line in lines]

        # Filter out lines containing certain strings or ending with certain characters
        filtered_lines = [line for line in masked_lines if not any(item in line for item in remove_lines_containing) and not line.endswith(('#', '>'))]

        return sorted(filtered_lines)

    def wait_until_tunnel_is_active(self, timeout: int = 30) -> None:
        count = 0
        for second in range(timeout, 0, -1):
            if count == 5:
                self.clear_crypto_sa()
            if self.is_tunnel_active:
                return
            sleep(1)
            logger.debug(f"Waiting for tunnel to appear ACTIVE ... {second}s/{timeout}s")
        raise TimeoutError("Timeout while waiting for tunnel to appear ACTIVE")

    def clear_crypto_sa(self) -> None:
        self.write_command("clear crypto isakmp")
        self.write_command("clear crypto ikev2 sa")
        self.write_command("clear crypto sa")

    def clear_crypto_counters(self) -> None:
        self.write_command("clear crypto counters")

    @get_response("show crypto acl")
    def parse_show_crypto_acl(self, response: Optional[str] = None) -> Optional[List[str]]:
        if len(response.split('\n')) <= 3:
            return ""
        else:
            return clean_lines(response, from_id=2, to_id=-2)

    @get_response("show policy-interface")
    def parse_show_policy_interface(self, response: Optional[str] = None) -> Optional[List[str]]:
        if len(response.split('\n')) <= 2:
            return ""
        else:
            return clean_lines(response, from_id=1, to_id=-1)

    def clear_policy_interface(self) -> None:
        self.write_command("clear policy-interface")

    @get_response("show license")
    def is_license_sdwanprime_activated(self, response: Optional[str] = None) -> bool:
        for line in response.split('\n'):
            if line.strip().startswith("sd-wan-prime"):
                installed = (line.split()[1] == "yes")
                activated = (line.split()[2] == "true")
                if not activated and installed:
                    logger.debug("SD WAN Prime license is installed but not activated")
                return activated
        raise AssertionError("SD WAN Prime license not found in 'show license' output")  # Should not happen

    @get_response("show license")
    def is_license_sdwanprime_installed(self, response: Optional[str] = None) -> bool:
        for line in response.split('\n'):
            if line.strip().startswith("sd-wan-prime"):
                return (line.split()[1] == "yes")
        raise AssertionError("SD WAN Prime license not found in 'show license' output")  # Should not happen


if __name__ == "__main__":
    pytest_args = ["-vvv", "--no-header", "--log-cli-level=DEBUG"]
    exit_code = pytest.main(pytest_args + ["-m ipsec"])
    sys.exit(exit_code)

    # Example of selection:
    # pytest.main(pytest_args + ["-m ipsec", "-k generic and setup1-"])
    # pytest.main(pytest_args + ["-m ipsec", "-k setup1-"])  # Will execute generic and alorithms setup1
    # pytest.main(pytest_args + ["-m ipsec", "-k generic and setup1"])  # Will execute setup1, setup10, setup11, ...
    # pytest.main(pytest_args + ["-m ipsec", "-k generic and not vm_c and not orange"])



# logger.py

"""
This module sets up a logger that outputs log messages to a file.

The logger is set to the DEBUG level, and outputs log messages to a file named debug.log in a directory named logs.
If the logs directory does not exist, it is created, but it should be ignored by git and only stored locally.

The log messages are formatted to include the date and time, the name of the logger, the level of the log message, and the log message itself.
Pytest automatically captures the log messages and prints them in the stdout.
"""

import os
import logging

def setup_logger():
    """
    Sets up a logger that outputs log messages to a file.

    This function creates a logger with the name of the current module, and sets its level to DEBUG.
    It also creates a file handler that outputs log messages to a file named debug.log in a directory named logs.
    If the logs directory does not exist, it is created.
    The log messages are formatted to include the date and time, the name of the logger, the level of the log message, and the log message itself.
    The file handler is added to the logger.
    """
    # Create a logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    # Create a logs directory if it doesn't exist
    if not os.path.exists("logs"):
        os.makedirs("logs")

    # Define a handler to output log messages to a file
    file_handler = logging.FileHandler("logs/debug.log")
    file_handler.setLevel(logging.DEBUG)

    # Define a formatter to output log messages with date and time
    formatter = logging.Formatter(
        "%(filename)s - %(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    file_handler.setFormatter(formatter)

    # Add file handler to the logger
    logger.addHandler(file_handler)



# device.py


"""
This module contains the Device base class and its subclasses.
The devices hold information specific to the device, such as the username, password and hostname.
Actions are performed to a connection to the device, which is held in the Connection class.
Actions do not originate from the device.
"""


import logging
import subprocess
from abc import ABC, abstractmethod
from typing import Optional


logger = logging.getLogger(__name__)


class Device(ABC):
    """
    Abstract base class for a device.

    Attributes:
        username (str): The username for the device. Default is None.
        password (str): The password for the device. Default is None.
        hostname (str): The hostname of the device. Default is None.
        _type (str): The type of the device. Default is "device".
    """

    def __init__(self, username: Optional[str]=None, password: Optional[str]=None):
        self.username = username
        self.password = password
        self.hostname = None
        self._type = "device"

    @property
    def type(self) -> str:
        return self._type

    @property
    @abstractmethod
    def DEFAULT_USERNAME(self):
        pass

    @property
    @abstractmethod
    def DEFAULT_PASSWORD(self):
        pass

    @property
    @abstractmethod
    def DEFAULT_PROMPT_SYMBOL(self):
        pass


class LinuxDevice(Device):
    DEFAULT_USERNAME = 'user'
    DEFAULT_PASSWORD = 'user'
    DEFAULT_PROMPT_SYMBOL = '$'  # Changes to '#' if root

    def __init__(self, username: Optional[str] = None, password: Optional[str] = None):
        username = username if username else self.DEFAULT_USERNAME
        password = password if password else self.DEFAULT_PASSWORD
        super().__init__(username, password)
        self._type = "linux"
        self.hostname = f"{self._type}-{username or self.DEFAULT_USERNAME}"


class RADIUSServer(LinuxDevice):
    def __init__(self, username: Optional[str]=None, password: Optional[str]=None):
        super().__init__(username, password)
        self.hostname = "radius-server"


class OneOS6Device(Device):
    DEFAULT_USERNAME = 'admin'
    DEFAULT_PASSWORD = 'admin'
    DEFAULT_PROMPT_SYMBOL = '#'

    # To be extended by use-case with more interfaces
    PHYSICAL_INTERFACES_LIST = [
        "gigabitethernet",
        "fastethernet",
    ]

    def __init__(self, username: Optional[str] = None, password: Optional[str] = None):
        username = username if username else self.DEFAULT_USERNAME
        password = password if password else self.DEFAULT_PASSWORD
        super().__init__(username, password)
        self._type = "oneos"
        self.hostname = "localhost"


class HostDevice():
    @staticmethod
    def write_command(command: str, print_response=False, quiet=False) -> Optional[str]:
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()

        responses = []
        if process.returncode != 0 and not quiet:
            logger.error(f"Error executing command: {command}")
            logger.error(f"Error message: {stderr.decode()}")
        else:
            response = stdout.decode()
            if print_response and not quiet:
                logger.debug(f"Command executed successfully: {command}")
                logger.debug(f"Output: {response}")
        responses.append(stdout.decode()) if stdout else None
        responses.append(stderr.decode()) if stderr else None
        return '\n'.join(responses) if responses else None



#!/usr/bin/env python3

"""
This is the setup configuration file for the virtual-router-test-kit package.
It uses setuptools to package the project, specify dependencies, and provide metadata.
"""

from setuptools import setup, find_packages


# Read the long description from README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Read the needed requirements from requirements.txt
with open("requirements.txt", "r", encoding="utf-8") as fh:
    install_requires = fh.read().splitlines()

setup(
    name="router-test-kit",
    version="0.1.2",
    author="Alexandros Anastasiou, alex-anast",
    author_email="anastasioyaa@gmail.com",
    description="A framework for remotely testing routers, physical or virtual.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alex-anast/router-test-kit",
    packages=find_packages(where="src"),  # Use 'src' as the package directory
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.7",
    install_requires=install_requires,
)
